#!/usr/bin/python

"""
*   Copyright (C) 2019 Ashim Ranjeet and Mukul S. Bansal (mukul.bansal@uconn.edu).
*
*   This program is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""



import argparse
from ete3 import Tree
import math

class LCAMapping:
    def __init__(self, Tree):
        self.tree = Tree
        self.array = []
        self.lcaToRMQ()
        self.preprocessST()
        

    # RMQ preprocessing using dynamic programming - O(n^2)
    def preprocessDP(self):
        n = len(self.levelLst)
        self.array = []
        for i in range(n):
            L1 = []
            for j in range(n):
                if(j < i):
                    L1.append('-')
                elif(j == i):
                    L1.append(i)
                else:
                    if(self.levelLst[L1[j-1]] > self.levelLst[j]):
                        L1.append(j)
                    else:
                        L1.append(L1[j-1])
            self.array.append(L1)
        #print(array)


    # RMQ preprocessing using Stable Table - O(nlogn)
    def preprocessST(self):
        self.array = []
        x = int(math.log2(len(self.levelLst)))
        for i in range(len(self.levelLst)):
            self.array.append([])
        for j in range(x+1):
            for i in range(len(self.levelLst)):
                if(j == 0):
                    self.array[i].append(i)
                else:
                    if(i+ 2**(j-1) >= len(self.levelLst)):
                        self.array[i].append('-')
                    elif(self.array[i+ 2**(j-1)][j-1] == '-'):
                        self.array[i].append('-')
                    elif(self.levelLst[self.array[i][j-1]] <= self.levelLst [ self.array[i+ 2**(j-1)][j-1] ]):
                        self.array[i].append(self.array[i][j-1])
                    else:
                        self.array[i].append(self.array[i+ 2**(j-1)][j-1])
    
        #return array


    
    
    
    # query for stable table prepreprossing
    def queryST(self, i, j):
        if(i == j ):
            if((i >= 0) and (i < len(self.levelLst)) ):
                return i
            else:
                print('out of bounds')
                return

        k = int(math.log2(j - i))
        x = self.array[i][k]
        y = self.array[j-2**(k)+1][k]
        if(self.levelLst[x] <= self.levelLst[y]):
            value = x
        else:
            value = y
        return value

        
    def eulerTour(self, node,lv):
        self.eTourLst.append(node)
        self.repLst[node.name] = len(self.eTourLst)-1
        self.levelLst.append(lv)
        if(node.is_leaf()):
            return True
        for nd in node.get_children():
            self.eulerTour(nd,lv+1)
            self.eTourLst.append(node)
            self.levelLst.append(lv)
        

    # convert LCA into RMQ
    def lcaToRMQ(self):
        self.eTourLst = []
        self.levelLst = []
        self.repLst = {}
        self.eulerTour(self.tree, 0)

    # query for LCA
    def queryLCA(self, node1, node2):
        x = self.repLst.get(node1)
        y = self.repLst.get(node2)
        if(x == y):
            indexL = x
        elif(x > y):
            indexL = self.queryST(y, x)
        elif(x < y):
            indexL = self.queryST(x, y)
        return self.eTourLst[indexL]



    

class compareTree:
    def __init__(self, T, S):
        self.colors = ["Green", "Red", "Blue"]
        self.t = T.copy()
        self.t2 = S.copy()
        
        self.iNode = 0      # if internal nodes are unnammed, name them based on numbers  
        self.swapped = 0    # to keep track of if t and t2 is T and S or if they are swapped
        self.resetT()
        

    def resetT(self):
        # swapping to set the smaller tree to be t, and larger tree to be t2
        # a map of the name to node based on dictionary is made.
        
        self.map = {}
        self.t1Leafset = {}
        self.t2Leafset = {}
        t1LeafCount = 0
        t2LeafCount = 0
        
        for node in self.t.traverse("postorder"):
            # if internal node doesn't have a name, name it inode, increment inode
            if(node.name == None or node.name == ''):    
                self.iNode = self.iNode + 1
                node.name = str(self.iNode)
            if(node.is_leaf()):
                t1LeafCount +=1
                self.t1Leafset[str(node.name)] = t1LeafCount

        for node in self.t2.traverse("postorder"):
            if(node.name == None or node.name == ''):    
                self.iNode = self.iNode + 1
                node.name = str(self.iNode)
            if(node.is_leaf()):
                t2LeafCount +=1
                self.t2Leafset[str(node.name)] = t2LeafCount

        if(t1LeafCount  > t2LeafCount):
            temp = self.t
            self.t = self.t2
            self.t2 = temp
            self.tLeafset = self.t2Leafset
            self.swapped = self.swapped + 1
        else:
            self.tLeafset = self.t1Leafset

        self.t1mapping = LCAMapping(self.t)
    
    def ROT_RF_Plus(self):
        #print(self.t.get_ascii(show_internal=True))
        #print(self.t2.get_ascii(show_internal=True))
        #tree.traverse("postorder") is ETE function which tranverse the tree in postorder
        for node in self.t2.traverse("postorder"):
            if(node.is_leaf()):     # coloring each leaf red or green
                if(self.tLeafset.get(str(node.name)) != None):
                    node.add_features(color = self.colors[0])
                    node.add_features(mark = False)
                else:
                    node.add_features(color = self.colors[1])
                    node.add_features(mark = False)
            else:   # coloring the internal nodes
                numGRB = [0,0,0]    # to count the number of green, red, blue children of the node
                # count the number of child depending on color
                for child in node.children:     
                    if(child.color == "Green"):
                        numGRB[0] = numGRB[0] + 1
                    elif(child.color == "Red"):
                        numGRB[1] = numGRB[1] + 1
                    elif(child.color == "Blue"):
                        numGRB[2] = numGRB[2] + 1
                # if both child are green, make color of node green.
                if(numGRB[0] == 2):
                    node.add_features(color = self.colors[0])
                    node.add_features(mark = False)
                # if both child are red, then make color of node red
                elif(numGRB[1] == 2):
                    node.add_features(color = self.colors[1])
                    node.add_features(mark = False)
                # if one child is red, and the other is green or blue, color node blue and mark node
                elif(numGRB[1] == 1 and (numGRB[0] == 1 or numGRB[2] == 1)):
                    node.add_features(color = self.colors[2])
                    node.add_features(mark = True)
                # otherwise mark node blue
                else:
                    node.add_features(color = self.colors[2])
                    node.add_features(mark = False)

        #ETE function to tranverse the tree in postorder (the second tree in this case)
        for node in self.t2.traverse("postorder"):
            # mapping
            # when the node is green
            if(node.color == self.colors[0]):
                if(node.is_leaf()):
                    # in the dictionary, map the name of the node to the node in the first tree
                    
                    self.map[node.name] = self.t1mapping.queryLCA(node.name, node.name)
                else:
                    # get a list of all the children, and find the common ancestor of the children
                    lst = []
                    for child in node.children:
                        lst.append((self.map).get(child.name))
                    # ETE function to get common ancestor of a list of leaves or nodes
                    #nd = self.t.get_common_ancestor(lst)
					# now geting the common ancestor using the LCA mapping
                    nd = self.t1mapping.queryLCA(lst[0].name, lst[1].name)
                    # map the name of the node to the ancestor node in the first tree 
                    self.map[node.name] = nd
            # when the node is blue
            elif(node.color == self.colors[2]):
                lst = []
                # and the children that are green or blue to the list
                for child in node.children:
                    if(child.color == self.colors[0] or child.color == self.colors[2]):
                        lst.append(self.map.get(child.name))
                # when there is only one node in the list, use that node is the common ancestor
                # otherwise get the greatest ancestor of the list of nodes in the first tree
                if(len(lst) == 1):
                    if(lst[0].is_leaf()):                        
                        nd = lst[0]
                    else:
                        #nd = self.t.get_common_ancestor(lst[0].children)
                        nd = self.t1mapping.queryLCA(lst[0].name, lst[0].name)
                    self.map[node.name] = nd
                elif(len(lst) == 2):
                    nd = self.t1mapping.queryLCA(lst[0].name,lst[1].name)
                    self.map[node.name] = nd
                else:
                    print('non binary tree')
        #ETE function to tranverse tree in preorder
        for node in self.t2.traverse("preorder"):   
            #tree-add
            # if the node is marked, meaning that one child is red
            if(node.mark == True):
                for child in node.children:
                    # when the child is red, add it that node to the first tree.
                    # if the node is the root, then the node is a sibling, to the current tree,
                    # and you have to create a new root
                    # otherwise, you have to detach the maping of the node in the first tree, and 
                    # create a new tree with the detach node and the node you want to add as sibling.
                    # this new tree is then added to location where the node was detached in the first tree.
                    if(child.color == self.colors[1]):
                        nd = self.map.get(node.name)
                        if(nd.up == None):
                            newT = Tree()
                            newT.add_child(self.t)
                            newT.add_child(child.copy())
                            self.t = newT
                        else:
                            up = nd.up
                            newT = Tree()
                            removed = nd.detach()
                            newT.add_child(child.copy())
                            newT.add_child(removed)
                            up.add_child(newT)
    
    def EF_R_RF(self):
        # calling the one tree completion function.
        # if there are missing leaves in both tree, then it is called twice
        # otherwise, it is only called once
        # it returns the two completed trees
        self.ROT_RF_Plus()
        self.resetT()
#        if(len(self.t1nodeMap) != len(self.t2nodeMap)):

        self.ROT_RF_Plus()
        if(self.swapped == 0 or self.swapped == 2):
            return(self.t, self.t2)
        else:
            return(self.t2, self.t)

def convertTreeIntoBinary(T):
    for node in T.traverse("postorder"):
        if ( not node.is_leaf() ):
            if (len(node.children) > 2):
                ndLst = []
                for child in node.children:
                    ndLst.append(child)
                for i in range(len(ndLst)):
                    ndLst[i] = ndLst[i].detach()
                oldTree = Tree()
                oldTree.add_child(ndLst[0])
                oldTree.add_child(ndLst[1])
                for i in range(2, len(ndLst)):
                    if( (i == len(ndLst) - 1) and (ndLst[i].up == None) ):
                        node.add_child(oldTree)
                        node.add_child(ndLst[i])
                    else:
                        newTree = Tree()
                        newTree.add_child(oldTree)
                        newTree.add_child(ndLst[i])
                        oldTree = newTree
            

        
        
def main():
    # takes an input file (required)
    # -o output file to print to is optional. when not present, print to treminal
    # -r is optional. when present, it prints out rf distance instead of modified trees
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--inputfile", type = str, help = "The input file contains the trees in newick format. This argument is required.")
    parser.add_argument("-o", "--outputfile", type = str, help = "The output file to which the rfdistance or trees in newick format will be printed")
    parser.add_argument("-r", "--rfdistance", action="store_true", help = "Type this command to print the RF(-) and RF(+) distances instead of the completed trees.")
    args = parser.parse_args()
    msn = []
    index = 0
    with open(args.inputfile) as openbn:
        for line in openbn:    
            x = line.strip(' ').strip('\t').strip('\n').strip(' ').strip('\t').strip('\n').strip(' ').strip('\t').strip('\n')
            if(x != ''):
                if(x[-1] != ";"):
                    x = x + ";"
                t = Tree(x)
                # checking to see if give tree is binary or not
                # with the new convertTreeIntoBinary, should no longer reach the error
                convertTreeIntoBinary(t)
                for node in t.traverse("preorder"):
                    if(len(node.children) > 2):
                        print("Error - non binary tree line: {}\n".format(index+1))
                        return
                msn.append(t)
                index += 1
    if(args.outputfile):
        fileNameW = args.outputfile
        fileW = open(fileNameW, "w")


    for i in range(1):
        for j in range(len(msn)):
            if(j > i):
                a = msn[i].robinson_foulds(msn[j])[0]
                y = compareTree(msn[i], msn[j]).EF_R_RF()
                z = y[0].robinson_foulds(y[1])[0]
                if(args.outputfile and args.rfdistance):
                   fileW.write("lines {} and {}: The RF(-) distance is {}. The RF(+) distance is {}.\n".format(i+1, j+1, a, z)) 
                elif(args.outputfile):
                    fileW.write(y[0].write(format = 9) + "\n")
                    fileW.write(y[1].write(format = 9) + "\n")
                elif(args.rfdistance):
                    print("lines {} and {}: The RF(-) distance is {}. The RF(+) distance is {}.".format(i+1, j+1, a, z))
                else:
                    print(y[0].write(format = 9))
                    print(y[1].write(format = 9) + "\n")  
    if(args.outputfile):
        fileW.close()



if __name__ == "__main__":
    main()
    
    
